@@ -2,10 +2,10 @@ use crate::{
     lock::Lock,
     manifest::{Dependency, Manifest},
 };
-use anyhow::{anyhow, bail, Result};
+use anyhow::{anyhow, bail, Context, Result};
 use forc_util::{
-    find_file_name, git_checkouts_directory, print_on_failure, print_on_success,
-    print_on_success_library, println_yellow_err,
+    find_file_name, git_checkouts_directory, kebab_to_snake_case, print_on_failure,
+    print_on_success, print_on_success_library, println_yellow_err,
 };
 use petgraph::{self, visit::EdgeRef, Directed, Direction};
 use serde::{Deserialize, Serialize};
@@ -20,11 +20,12 @@ use sway_core::{
     NamespaceWrapper, TreeType, TypedParseTree,
 };
 use sway_types::JsonABI;
+use sway_utils::constants;
 use url::Url;
 
 type GraphIx = u32;
 type Node = Pinned;
-type Edge = ();
+type Edge = DependencyName;
 pub type Graph = petgraph::Graph<Node, Edge, Directed, GraphIx>;
 pub type NodeIx = petgraph::graph::NodeIndex<GraphIx>;
 pub type PathMap = HashMap<PinnedId, PathBuf>;
@@ -44,7 +45,7 @@ pub struct Compiled {
 /// A package uniquely identified by name along with its source.
 #[derive(Clone, Debug, Eq, Hash, Ord, PartialEq, PartialOrd, Deserialize, Serialize)]
 pub struct Pkg {
-    /// The unique name of the package.
+    /// The unique name of the package as declared in its manifest.
     pub name: String,
     /// Where the package is sourced from.
     pub source: Source,
@@ -80,7 +81,19 @@ pub struct SourceGit {
     /// The URL at which the repository is located.
     pub repo: Url,
     /// A git reference, e.g. a branch or tag.
-    pub reference: String,
+    pub reference: GitReference,
+}
+
+/// Used to distinguish between types of git references.
+///
+/// For the most part, `GitReference` is useful to refine the `refspecs` used to fetch remote
+/// repositories.
+#[derive(Clone, Debug, Eq, Hash, Ord, PartialEq, PartialOrd, Deserialize, Serialize)]
+pub enum GitReference {
+    Branch(String),
+    Tag(String),
+    Rev(String),
+    DefaultBranch,
 }
 
 /// A package from the official registry.
@@ -145,6 +158,26 @@ pub enum SourceGitPinnedParseError {
     CommitHash,
 }
 
+/// The name specified on the left hand side of the `=` in a depenedency declaration under
+/// `[dependencies]` within a forc manifest.
+///
+/// The name of a dependency may differ from the package name in the case that the dependency's
+/// `package` field is specified.
+///
+/// For example, in the following, `foo` is assumed to be both the package name and the dependency
+/// name:
+///
+/// ```toml
+/// foo = { git = "https://github.com/owner/repo", branch = "master" }
+/// ```
+///
+/// In the following case however, `foo` is the package name, but the dependency name is `foo-alt`:
+///
+/// ```toml
+/// foo-alt = { git = "https://github.com/owner/repo", branch = "master", package = "foo" }
+/// ```
+pub type DependencyName = String;
+
 impl BuildPlan {
     /// Create a new build plan for the project by fetching and pinning dependenies.
     pub fn new(manifest_dir: &Path, offline: bool) -> Result<Self> {
@@ -189,7 +222,11 @@ impl BuildPlan {
         let plan_dep_pkgs: BTreeSet<_> = self
             .graph
             .edges_directed(proj_node, Direction::Outgoing)
-            .map(|e| self.graph[e.target()].unpinned(&self.path_map))
+            .map(|e| {
+                let dep_name = e.weight();
+                let dep_pkg = self.graph[e.target()].unpinned(&self.path_map);
+                (dep_name, dep_pkg)
+            })
             .collect();
 
         // Collect dependency `Source`s from manifest.
@@ -200,22 +237,23 @@ impl BuildPlan {
             .as_ref()
             .into_iter()
             .flat_map(|deps| deps.iter())
-            .map(|(name, dep)| {
+            .map(|(dep_name, dep)| {
                 // NOTE: Temporarily warn about `version` until we have support for registries.
                 if let Dependency::Detailed(det) = dep {
                     if det.version.is_some() {
                         println_yellow_err(&format!(
                             "  WARNING! Dependency \"{}\" specifies the unused `version` field: \
                             consider using `branch` or `tag` instead",
-                            name
+                            dep_name
                         ))
                         .unwrap();
                     }
                 }
 
-                let name = name.clone();
+                let name = dep.package().unwrap_or(dep_name).to_string();
                 let source = dep_to_source(proj_path, dep)?;
-                Ok(Pkg { name, source })
+                let dep_pkg = Pkg { name, source };
+                Ok((dep_name, dep_pkg))
             })
             .collect::<Result<BTreeSet<_>>>()?;
 
@@ -224,6 +262,21 @@ impl BuildPlan {
             bail!("Manifest dependencies do not match");
         }
 
+        // Ensure the pkg names of all nodes match their associated manifests.
+        for node in self.graph.node_indices() {
+            let pkg = &self.graph[node];
+            let id = pkg.id();
+            let path = &self.path_map[&id];
+            let manifest = Manifest::from_dir(path)?;
+            if pkg.name != manifest.project.name {
+                bail!(
+                    "package name {:?} does not match the associated manifest project name {:?}",
+                    pkg.name,
+                    manifest.project.name,
+                );
+            }
+        }
+
         Ok(())
     }
 
@@ -244,6 +297,56 @@ impl BuildPlan {
     }
 }
 
+impl GitReference {
+    /// Resolves the parsed forc git reference to the associated git ID.
+    pub fn resolve(&self, repo: &git2::Repository) -> Result<git2::Oid> {
+        // Find the commit associated with this tag.
+        fn resolve_tag(repo: &git2::Repository, tag: &str) -> Result<git2::Oid> {
+            let refname = format!("refs/remotes/origin/tags/{}", tag);
+            let id = repo.refname_to_id(&refname)?;
+            let obj = repo.find_object(id, None)?;
+            let obj = obj.peel(git2::ObjectType::Commit)?;
+            Ok(obj.id())
+        }
+
+        // Resolve to the target for the given branch.
+        fn resolve_branch(repo: &git2::Repository, branch: &str) -> Result<git2::Oid> {
+            let name = format!("origin/{}", branch);
+            let b = repo
+                .find_branch(&name, git2::BranchType::Remote)
+                .with_context(|| format!("failed to find branch `{}`", branch))?;
+            b.get()
+                .target()
+                .ok_or_else(|| anyhow::format_err!("branch `{}` did not have a target", branch))
+        }
+
+        // Use the HEAD commit when default branch is specified.
+        fn resolve_default_branch(repo: &git2::Repository) -> Result<git2::Oid> {
+            let head_id = repo.refname_to_id("refs/remotes/origin/HEAD")?;
+            let head = repo.find_object(head_id, None)?;
+            Ok(head.peel(git2::ObjectType::Commit)?.id())
+        }
+
+        // Find the commit for the given revision.
+        fn resolve_rev(repo: &git2::Repository, rev: &str) -> Result<git2::Oid> {
+            let obj = repo.revparse_single(rev)?;
+            match obj.as_tag() {
+                Some(tag) => Ok(tag.target_id()),
+                None => Ok(obj.id()),
+            }
+        }
+
+        match self {
+            GitReference::Tag(s) => {
+                resolve_tag(repo, s).with_context(|| format!("failed to find tag `{}`", s))
+            }
+            GitReference::Branch(s) => resolve_branch(repo, s),
+            GitReference::DefaultBranch => resolve_default_branch(repo),
+            GitReference::Rev(s) => resolve_rev(repo, s),
+        }
+    }
+}
+
 impl Pinned {
     /// Retrieve the unique ID for the pinned package.
     ///
@@ -277,18 +380,30 @@ impl PinnedId {
 
 impl ToString for SourceGitPinned {
     fn to_string(&self) -> String {
-        // git+<url/to/repo>?reference=<reference>#<commit>
+        // git+<url/to/repo>?<ref_kind>=<ref_string>#<commit>
+        let reference = self.source.reference.to_string();
         format!(
-            "git+{}?reference={}#{}",
-            self.source.repo, self.source.reference, self.commit_hash,
+            "git+{}?{}#{}",
+            self.source.repo, reference, self.commit_hash
         )
     }
 }
 
+impl ToString for GitReference {
+    fn to_string(&self) -> String {
+        match self {
+            GitReference::Branch(ref s) => format!("branch={}", s),
+            GitReference::Tag(ref s) => format!("tag={}", s),
+            GitReference::Rev(ref _s) => "rev".to_string(),
+            GitReference::DefaultBranch => "default-branch".to_string(),
+        }
+    }
+}
+
 impl FromStr for SourceGitPinned {
     type Err = SourceGitPinnedParseError;
     fn from_str(s: &str) -> Result<Self, Self::Err> {
-        // git+<url/to/repo>?reference=<reference>#<commit>
+        // git+<url/to/repo>?<reference>#<commit>
         let s = s.trim();
 
         // Check for "git+" at the start.
@@ -303,25 +418,33 @@ impl FromStr for SourceGitPinned {
         let repo = Url::parse(repo_str).map_err(|_| SourceGitPinnedParseError::Url)?;
         let s = &s[repo_str.len() + "?".len()..];
 
-        // Parse the "reference=" string.
-        // TODO: This will need updating if we want to support omitting a git reference and allow
-        // for specifying commit hashes directly in `Forc.toml` git dependencies.
-        const REFERENCE: &str = "reference=";
-        if s.find(REFERENCE) != Some(0) {
-            return Err(SourceGitPinnedParseError::Reference);
-        }
-        let s = &s[REFERENCE.len()..];
-
-        // And now retrieve the `reference` and `commit_hash` values.
+        // Parse the git reference and commit hash. This can be any of either:
+        // - `branch=<branch-name>#<commit-hash>`
+        // - `tag=<tag-name>#<commit-hash>`
+        // - `rev#<commit-hash>`
+        // - `default#<commit-hash>`
         let mut s_iter = s.split('#');
-        let reference = s_iter
-            .next()
-            .ok_or(SourceGitPinnedParseError::Reference)?
-            .to_string();
+        let reference = s_iter.next().ok_or(SourceGitPinnedParseError::Reference)?;
         let commit_hash = s_iter
             .next()
             .ok_or(SourceGitPinnedParseError::CommitHash)?
             .to_string();
+        validate_git_commit_hash(&commit_hash)
+            .map_err(|_| SourceGitPinnedParseError::CommitHash)?;
+
+        const BRANCH: &str = "branch=";
+        const TAG: &str = "tag=";
+        let reference = if reference.find(BRANCH) == Some(0) {
+            GitReference::Branch(reference[BRANCH.len()..].to_string())
+        } else if reference.find(TAG) == Some(0) {
+            GitReference::Tag(reference[TAG.len()..].to_string())
+        } else if reference == "rev" {
+            GitReference::Rev(commit_hash.to_string())
+        } else if reference == "default-branch" {
+            GitReference::DefaultBranch
+        } else {
+            return Err(SourceGitPinnedParseError::Reference);
+        };
 
         let source = SourceGit { repo, reference };
         Ok(Self {
@@ -331,6 +454,27 @@ impl FromStr for SourceGitPinned {
     }
 }
 
+fn validate_git_commit_hash(commit_hash: &str) -> Result<()> {
+    const LEN: usize = 40;
+    if commit_hash.len() != LEN {
+        bail!(
+            "invalid hash length: expected {}, found {}",
+            LEN,
+            commit_hash.len()
+        );
+    }
+    if !commit_hash.chars().all(|c| c.is_ascii_alphanumeric()) {
+        bail!("hash contains one or more non-ascii-alphanumeric characters");
+    }
+    Ok(())
+}
+
+impl Default for GitReference {
+    fn default() -> Self {
+        Self::DefaultBranch
+    }
+}
+
 /// The `pkg::Graph` is of *a -> b* where *a* depends on *b*. We can determine compilation order by
 /// performing a toposort of the graph with reversed weights. The resulting order ensures all
 /// dependencies are always compiled before their dependents.
@@ -370,7 +514,18 @@ pub fn graph_to_path_map(
         let dep = &graph[dep_node];
         let dep_path = match &dep.source {
             SourcePinned::Git(git) => {
-                git_commit_path(&dep.name, &git.source.repo, &git.commit_hash)
+                let repo_path = git_commit_path(&dep.name, &git.source.repo, &git.commit_hash);
+                if !repo_path.exists() {
+                    println!("  Fetching {}", git.to_string());
+                    fetch_git(fetch_id, &dep.name, git)?;
+                }
+                find_dir_within(&repo_path, &dep.name).ok_or_else(|| {
+                    anyhow!(
+                        "failed to find package `{}` in {}",
+                        dep.name,
+                        git.to_string()
+                    )
+                })?
             }
             SourcePinned::Path => {
                 let parent_node = graph
@@ -393,26 +548,16 @@ pub fn graph_to_path_map(
                     .path
                     .as_ref()
                     .ok_or_else(|| anyhow!("missing path info for dependency: {}", dep.name))?;
-                parent_path.join(rel_dep_path)
+                let path = parent_path.join(rel_dep_path);
+                if !path.exists() {
+                    bail!("pinned `path` dependency \"{}\" source missing", dep.name);
+                }
+                path
             }
             SourcePinned::Registry(_reg) => {
                 bail!("registry dependencies are not yet supported");
             }
         };
-        if !dep_path.exists() {
-            match &dep.source {
-                SourcePinned::Path => {
-                    bail!("pinned `path` dependency \"{}\" source missing", dep.name);
-                }
-                SourcePinned::Git(git) => {
-                    println!("  Fetching {}", git.to_string());
-                    fetch_git(fetch_id, &dep.name, git)?;
-                }
-                SourcePinned::Registry(_reg) => {
-                    bail!("registry dependencies are not yet supported");
-                }
-            }
-        }
         path_map.insert(dep.id(), dep_path);
     }
 
@@ -448,10 +593,12 @@ pub(crate) fn fetch_deps(
     // TODO: Convert this recursion to use loop & stack to ensure deps can't cause stack overflow.
     let fetch_ts = std::time::Instant::now();
     let fetch_id = fetch_id(&path_map[&pkg_id], fetch_ts);
+    let manifest = Manifest::from_dir(&path_map[&pkg_id])?;
     fetch_children(
         fetch_id,
         offline_mode,
         root,
+        &manifest,
         &mut graph,
         &mut path_map,
         &mut visited,
@@ -475,34 +622,49 @@ fn fetch_children(
     fetch_id: u64,
     offline_mode: bool,
     node: NodeIx,
+    manifest: &Manifest,
     graph: &mut Graph,
     path_map: &mut PathMap,
     visited: &mut HashMap<Pinned, NodeIx>,
 ) -> Result<()> {
     let parent = &graph[node];
     let parent_path = path_map[&parent.id()].clone();
-    let manifest = Manifest::from_dir(&parent_path)?;
-    let deps = match &manifest.dependencies {
-        None => return Ok(()),
-        Some(deps) => deps,
-    };
-    for (name, dep) in deps {
-        let name = name.clone();
+    for (dep_name, dep) in manifest.deps() {
+        let name = dep.package().unwrap_or(dep_name).to_string();
         let source = dep_to_source(&parent_path, dep)?;
         if offline_mode && !matches!(source, Source::Path(_)) {
             bail!("Unable to fetch pkg {:?} in offline mode", source);
         }
         let pkg = Pkg { name, source };
         let pinned = pin_pkg(fetch_id, &pkg, path_map)?;
+        let pkg_id = pinned.id();
+        let manifest = Manifest::from_dir(&path_map[&pkg_id])?;
+        if pinned.name != manifest.project.name {
+            bail!(
+                "dependency name {:?} must match the manifest project name {:?} \
+                unless `package = {:?}` is specified in the dependency declaration",
+                pinned.name,
+                manifest.project.name,
+                manifest.project.name,
+            );
+        }
         let dep_node = if let hash_map::Entry::Vacant(entry) = visited.entry(pinned.clone()) {
             let node = graph.add_node(pinned);
             entry.insert(node);
-            fetch_children(fetch_id, offline_mode, node, graph, path_map, visited)?;
+            fetch_children(
+                fetch_id,
+                offline_mode,
+                node,
+                &manifest,
+                graph,
+                path_map,
+                visited,
+            )?;
             node
         } else {
             visited[&pinned]
         };
-        graph.add_edge(node, dep_node, ());
+        graph.add_edge(node, dep_node, dep_name.to_string());
     }
     Ok(())
 }
@@ -536,36 +698,70 @@ fn tmp_git_repo_dir(fetch_id: u64, name: &str, repo: &Url) -> PathBuf {
     git_checkouts_directory().join("tmp").join(repo_dir_name)
 }
 
-/// Clones the package git repo into a temporary directory and applies the given function.
-fn with_tmp_git_repo<F, O>(fetch_id: u64, name: &str, source: &Url, f: F) -> Result<O>
+/// Given a git reference, build a list of `refspecs` required for the fetch opration.
+///
+/// Also returns whether or not our reference implies we require fetching tags.
+fn git_ref_to_refspecs(reference: &GitReference) -> (Vec<String>, bool) {
+    let mut refspecs = vec![];
+    let mut tags = false;
+    match reference {
+        GitReference::Branch(s) => {
+            refspecs.push(format!("+refs/heads/{0}:refs/remotes/origin/{0}", s));
+        }
+        GitReference::Tag(s) => {
+            refspecs.push(format!("+refs/tags/{0}:refs/remotes/origin/tags/{0}", s));
+        }
+        GitReference::Rev(s) => {
+            if s.starts_with("refs/") {
+                refspecs.push(format!("+{0}:{0}", s));
+            } else {
+                // We can't fetch the commit directly, so we fetch all branches and tags in order
+                // to find it.
+                refspecs.push("+refs/heads/*:refs/remotes/origin/*".to_string());
+                refspecs.push("+HEAD:refs/remotes/origin/HEAD".to_string());
+                tags = true;
+            }
+        }
+        GitReference::DefaultBranch => {
+            refspecs.push("+HEAD:refs/remotes/origin/HEAD".to_string());
+        }
+    }
+    (refspecs, tags)
+}
+
+/// Initializes a temporary git repo for the package and fetches only the reference associated with
+/// the given source.
+fn with_tmp_git_repo<F, O>(fetch_id: u64, name: &str, source: &SourceGit, f: F) -> Result<O>
 where
     F: FnOnce(git2::Repository) -> Result<O>,
 {
     // Clear existing temporary directory if it exists.
-    let repo_dir = tmp_git_repo_dir(fetch_id, name, source);
+    let repo_dir = tmp_git_repo_dir(fetch_id, name, &source.repo);
     if repo_dir.exists() {
         let _ = std::fs::remove_dir_all(&repo_dir);
     }
 
-    // Clone repo into temporary directory.
-    let repo_url_string = format!("{}", source);
-    let repo = git2::Repository::clone(&repo_url_string, &repo_dir).map_err(|e| {
-        anyhow!(
-            "failed to clone package '{}' from '{}': {}",
-            name,
-            source,
-            e
-        )
-    })?;
+    // Initialise the repository.
+    let repo = git2::Repository::init(&repo_dir)
+        .map_err(|e| anyhow!("failed to init repo at \"{}\": {}", repo_dir.display(), e))?;
 
-    // Do something with the repo.
+    // Fetch the necessary references.
+    let (refspecs, tags) = git_ref_to_refspecs(&source.reference);
+
+    // Fetch the refspecs.
+    let mut fetch_opts = git2::FetchOptions::new();
+    if tags {
+        fetch_opts.download_tags(git2::AutotagOption::All);
+    }
+    repo.remote_anonymous(source.repo.as_str())?
+        .fetch(&refspecs, Some(&mut fetch_opts), None)
+        .with_context(|| format!("failed to fetch `{}`", &source.repo))?;
+
+    // Call the user function.
     let output = f(repo)?;
 
     // Clean up the temporary directory.
-    if repo_dir.exists() {
-        let _ = std::fs::remove_dir_all(&repo_dir);
-    }
-
+    let _ = std::fs::remove_dir_all(&repo_dir);
     Ok(output)
 }
 
@@ -574,29 +770,13 @@ where
 /// This clones the repository to a temporary directory in order to determine the commit at the
 /// HEAD of the given git reference.
 fn pin_git(fetch_id: u64, name: &str, source: SourceGit) -> Result<SourceGitPinned> {
-    let commit_hash = with_tmp_git_repo(fetch_id, name, &source.repo, |repo| {
-        // Find specified reference in repo.
-        let reference = repo
-            .resolve_reference_from_short_name(&source.reference)
-            .map_err(|e| {
-                anyhow!(
-                    "failed to find git ref '{}' for package '{}': {}",
-                    source.reference,
-                    name,
-                    e
-                )
-            })?;
-
-        // Follow the reference until we find the latest commit and retrieve its hash.
-        let commit = reference.peel_to_commit().map_err(|e| {
-            anyhow!(
-                "failed to obtain commit for ref '{}' of package '{}': {}",
-                source.reference,
-                name,
-                e
-            )
-        })?;
-        Ok(format!("{}", commit.id()))
+    let commit_hash = with_tmp_git_repo(fetch_id, name, &source, |repo| {
+        // Resolve the reference to the commit ID.
+        let commit_id = source
+            .reference
+            .resolve(&repo)
+            .with_context(|| "failed to resolve reference".to_string())?;
+        Ok(format!("{}", commit_id))
     })?;
     Ok(SourceGitPinned {
         source,
@@ -617,9 +797,10 @@ fn pin_pkg(fetch_id: u64, pkg: &Pkg, path_map: &mut PathMap) -> Result<Pinned> {
             path_map.insert(id, path.clone());
             pinned
         }
-        Source::Git(ref source) => {
-            let pinned_git = pin_git(fetch_id, &name, source.clone())?;
-            let path = git_commit_path(&name, &pinned_git.source.repo, &pinned_git.commit_hash);
+        Source::Git(ref git_source) => {
+            let pinned_git = pin_git(fetch_id, &name, git_source.clone())?;
+            let repo_path =
+                git_commit_path(&name, &pinned_git.source.repo, &pinned_git.commit_hash);
             let source = SourcePinned::Git(pinned_git.clone());
             let pinned = Pinned { name, source };
             let id = pinned.id();
@@ -629,10 +810,17 @@ fn pin_pkg(fetch_id: u64, pkg: &Pkg, path_map: &mut PathMap) -> Result<Pinned> {
                 // cases as users should never be touching these directories, however we should add some code
                 // to validate this. E.g. can we recreate the git hash by hashing the directory or something
                 // along these lines using git?
-                if !path.exists() {
+                if !repo_path.exists() {
                     println!("  Fetching {}", pinned_git.to_string());
                     fetch_git(fetch_id, &pinned.name, &pinned_git)?;
                 }
+                let path = find_dir_within(&repo_path, &pinned.name).ok_or_else(|| {
+                    anyhow!(
+                        "failed to find package `{}` in {}",
+                        pinned.name,
+                        pinned_git.to_string()
+                    )
+                })?;
                 entry.insert(path);
             }
             pinned
@@ -669,7 +857,7 @@ fn fetch_git(fetch_id: u64, name: &str, pinned: &SourceGitPinned) -> Result<Path
     let path = git_commit_path(name, &pinned.source.repo, &pinned.commit_hash);
 
     // Checkout the pinned hash to the path.
-    with_tmp_git_repo(fetch_id, name, &pinned.source.repo, |repo| {
+    with_tmp_git_repo(fetch_id, name, &pinned.source, |repo| {
         // Change HEAD to point to the pinned commit.
         let id = git2::Oid::from_str(&pinned.commit_hash)?;
         repo.set_head_detached(id)?;
@@ -693,29 +881,38 @@ fn fetch_git(fetch_id: u64, name: &str, pinned: &SourceGitPinned) -> Result<Path
 /// produce the `Source` for that dependendency.
 fn dep_to_source(pkg_path: &Path, dep: &Dependency) -> Result<Source> {
     let source = match dep {
-        Dependency::Simple(ref _ver_str) => unimplemented!(),
-        Dependency::Detailed(ref det) => {
-            match (&det.path, &det.version, &det.git, &det.branch, &det.tag) {
-                (Some(relative_path), _, _, _, _) => {
-                    let path = pkg_path.join(relative_path);
-                    Source::Path(path)
-                }
-                (_, _, Some(repo), branch, tag) => {
-                    let reference = match (branch, tag) {
-                        (Some(branch), None) => branch.clone(),
-                        (None, Some(tag)) => tag.clone(),
-                        // TODO: Consider "main" or having no default at all.
-                        _ => "master".to_string(),
-                    };
-                    let repo = Url::parse(repo)?;
-                    let source = SourceGit { repo, reference };
-                    Source::Git(source)
-                }
-                _ => {
-                    bail!("unsupported set of arguments for dependency: {:?}", dep);
-                }
-            }
+        Dependency::Simple(ref ver_str) => {
+            bail!(
+                "Unsupported dependency declaration in \"{}\": `{}` - \
+                currently only `git` and `path` dependencies are supported",
+                pkg_path.display(),
+                ver_str
+            )
         }
+        Dependency::Detailed(ref det) => match (&det.path, &det.version, &det.git) {
+            (Some(relative_path), _, _) => {
+                let path = pkg_path.join(relative_path);
+                Source::Path(path)
+            }
+            (_, _, Some(repo)) => {
+                let reference = match (&det.branch, &det.tag, &det.rev) {
+                    (Some(branch), None, None) => GitReference::Branch(branch.clone()),
+                    (None, Some(tag), None) => GitReference::Tag(tag.clone()),
+                    (None, None, Some(rev)) => GitReference::Rev(rev.clone()),
+                    (None, None, None) => GitReference::DefaultBranch,
+                    _ => bail!(
+                        "git dependencies support at most one reference: \
+                            either `branch`, `tag` or `rev`"
+                    ),
+                };
+                let repo = Url::parse(repo)?;
+                let source = SourceGit { repo, reference };
+                Source::Git(source)
+            }
+            _ => {
+                bail!("unsupported set of fields for dependency: {:?}", dep);
+            }
+        },
     };
     Ok(source)
 }
@@ -757,11 +954,20 @@ pub fn dependency_namespace(
 
     // In order of compilation, accumulate dependency namespace refs.
     let namespace = sway_core::create_module();
-    for dep_node in compilation_order.iter().filter(|n| deps.contains(n)) {
-        if *dep_node == node {
+    for &dep_node in compilation_order.iter().filter(|n| deps.contains(n)) {
+        if dep_node == node {
             break;
         }
-        namespace.insert_module_ref(graph[*dep_node].name.clone(), namespace_map[dep_node]);
+        // Add the namespace once for each of its names.
+        let namespace_ref = namespace_map[&dep_node];
+        let dep_names: BTreeSet<_> = graph
+            .edges_directed(dep_node, Direction::Incoming)
+            .map(|e| e.weight())
+            .collect();
+        for dep_name in dep_names {
+            let dep_name = kebab_to_snake_case(dep_name);
+            namespace.insert_module_ref(dep_name.to_string(), namespace_ref);
+        }
     }
 
     namespace
@@ -875,6 +1081,30 @@ pub fn build(plan: &BuildPlan, conf: &BuildConfig) -> anyhow::Result<(Compiled,
     Ok((compiled, source_map))
 }
 
+/// Attempt to find a `Forc.toml` with the given project name within the given directory.
+///
+/// Returns the path to the package on success, or `None` in the case it could not be found.
+pub fn find_within(dir: &Path, pkg_name: &str) -> Option<PathBuf> {
+    walkdir::WalkDir::new(dir)
+        .into_iter()
+        .filter_map(Result::ok)
+        .filter(|entry| entry.path().ends_with(constants::MANIFEST_FILE_NAME))
+        .find_map(|entry| {
+            let path = entry.path();
+            let manifest = Manifest::from_file(path).ok()?;
+            if manifest.project.name == pkg_name {
+                Some(path.to_path_buf())
+            } else {
+                None
+            }
+        })
+}
+
+/// The same as [find_within], but returns the package's project directory.
+pub fn find_dir_within(dir: &Path, pkg_name: &str) -> Option<PathBuf> {
+    find_within(dir, pkg_name).and_then(|path| path.parent().map(Path::to_path_buf))
+}
+
 // TODO: Update this to match behaviour described in the `compile` doc comment above.
 fn generate_json_abi(ast: &TypedParseTree) -> JsonABI {
     match ast {
@@ -888,33 +1118,51 @@ fn generate_json_abi(ast: &TypedParseTree) -> JsonABI {
 #[test]
 fn test_source_git_pinned_parsing() {
     let strings = [
-        "git+https://github.com/foo/bar?reference=baz#64092602dd6158f3e41d775ed889389440a2cd86",
-        "git+https://github.com/fuellabs/sway-lib-std?reference=v0.1.0#0000000000000000000000000000000000000000",
-        "git+https://github.com/fuellabs/sway-lib-core?reference=v0.0.1#0000000000000000000000000000000000000000",
+        "git+https://github.com/foo/bar?branch=baz#64092602dd6158f3e41d775ed889389440a2cd86",
+        "git+https://github.com/fuellabs/sway-lib-std?tag=v0.1.0#0000000000000000000000000000000000000000",
+        "git+https://github.com/fuellabs/sway-lib-core?tag=v0.0.1#0000000000000000000000000000000000000000",
+        "git+https://some-git-host.com/owner/repo?rev#FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF",
+        "git+https://some-git-host.com/owner/repo?default-branch#AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA",
     ];
 
     let expected = [
         SourceGitPinned {
             source: SourceGit {
                 repo: Url::parse("https://github.com/foo/bar").unwrap(),
-                reference: "baz".to_string(),
+                reference: GitReference::Branch("baz".to_string()),
             },
             commit_hash: "64092602dd6158f3e41d775ed889389440a2cd86".to_string(),
         },
         SourceGitPinned {
             source: SourceGit {
                 repo: Url::parse("https://github.com/fuellabs/sway-lib-std").unwrap(),
-                reference: "v0.1.0".to_string(),
+                reference: GitReference::Tag("v0.1.0".to_string()),
             },
             commit_hash: "0000000000000000000000000000000000000000".to_string(),
         },
         SourceGitPinned {
             source: SourceGit {
                 repo: Url::parse("https://github.com/fuellabs/sway-lib-core").unwrap(),
-                reference: "v0.0.1".to_string(),
+                reference: GitReference::Tag("v0.0.1".to_string()),
             },
             commit_hash: "0000000000000000000000000000000000000000".to_string(),
         },
+        SourceGitPinned {
+            source: SourceGit {
+                repo: Url::parse("https://some-git-host.com/owner/repo").unwrap(),
+                reference: GitReference::Rev(
+                    "FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF".to_string(),
+                ),
+            },
+            commit_hash: "FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF".to_string(),
+        },
+        SourceGitPinned {
+            source: SourceGit {
+                repo: Url::parse("https://some-git-host.com/owner/repo").unwrap(),
+                reference: GitReference::DefaultBranch,
+            },
+            commit_hash: "AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA".to_string(),
+        },
     ];
 
     for (&string, expected) in strings.iter().zip(&expected) {
